import warnings
from typing import (
    Iterable,
    Optional,
)

from ..abc import (
    EnrouteDecorator,
)


class HttpEnrouteDecorator(EnrouteDecorator):
    """Http Enroute Decorator class."""

    def __init__(self, path: Optional[str] = None, method: Optional[str] = None, url: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        if path is None and url is not None:
            warnings.warn("The 'url' argument has been deprecated. 'path' must be used.", DeprecationWarning)
            path = url
        if path is None:
            raise ValueError("A 'path' must be provided.")
        if method is None:
            raise ValueError("A 'method' must be provided.")
        self.path = path
        self.method = method

    def __iter__(self) -> Iterable:
        yield from (
            self.path,
            self.method,
        )

    @property
    def url(self) -> str:
        """Get the url.

        :return: A ``str`` value.
        """
        return self.path



from .contextvars import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
)
from .models import (
    BrokerMessage,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerMessageV1Status,
    BrokerMessageV1Strategy,
)



from minos.common import (
    Builder,
)

from ...messages import (
    BrokerMessage,
)
from ..abc import (
    BrokerSubscriber,
)
from .validators import (
    BrokerSubscriberValidator,
)

_sentinel = object()


class FilteredBrokerSubscriber(BrokerSubscriber):
    """Filtered Broker Subscriber class."""

    impl: BrokerSubscriber
    validator: BrokerSubscriberValidator

    def __init__(self, impl: BrokerSubscriber, validator: BrokerSubscriberValidator, **kwargs):
        super().__init__(**(kwargs | {"topics": impl.topics}))
        self.impl = impl
        self.validator = validator

    async def _setup(self) -> None:
        await super()._setup()
        await self.validator.setup()
        await self.impl.setup()

    async def _destroy(self) -> None:
        await self.impl.destroy()
        await self.validator.destroy()
        await super()._destroy()

    async def _receive(self) -> BrokerMessage:
        message = _sentinel
        while message is _sentinel or not (await self.validator.is_valid(message)):
            message = await self.impl.receive()
        return message


FilteredBrokerSubscriber.set_builder(Builder)



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    CancelledError,
    Event,
    PriorityQueue,
    QueueEmpty,
    TimeoutError,
    create_task,
    wait_for,
)
from contextlib import (
    suppress,
)
from typing import (
    Any,
    Generic,
    NoReturn,
    Optional,
    TypeVar,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Builder,
    Config,
    DatabaseClient,
    DatabaseMixin,
)

from ....messages import (
    BrokerMessage,
)
from ..abc import (
    BrokerQueue,
)
from .factories import (
    BrokerQueueDatabaseOperationFactory,
)

logger = logging.getLogger(__name__)

GenericBrokerQueueDatabaseOperationFactory = TypeVar(
    "GenericBrokerQueueDatabaseOperationFactory", bound=BrokerQueueDatabaseOperationFactory
)


class DatabaseBrokerQueue(
    BrokerQueue,
    DatabaseMixin[GenericBrokerQueueDatabaseOperationFactory],
    Generic[GenericBrokerQueueDatabaseOperationFactory],
):
    """Database Broker Queue class."""

    _queue: PriorityQueue[_Entry]

    def __init__(
        self,
        *args,
        retry: Optional[int] = None,
        records: Optional[int] = None,
        database_key: Optional[tuple[str]] = None,
        **kwargs,
    ):
        if database_key is None:
            database_key = ("broker",)
        super().__init__(*args, database_key=database_key, **kwargs)

        if retry is None:
            retry = 2
        if records is None:
            records = 1000

        self._retry = retry
        self._records = records

        self._queue = PriorityQueue(maxsize=records)

        self._run_task = None
        self._enqueued_event = Event()

    @property
    def retry(self) -> int:
        """Get the retry value.

        :return: A ``int`` value.
        """
        return self._retry

    @property
    def records(self) -> int:
        """Get the records value.

        :return: A ``int`` value.
        """
        return self._records

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> DatabaseBrokerQueue:
        broker_interface = config.get_interface_by_name("broker")
        queue_config = broker_interface.get("common", dict()).get("queue", dict())
        database_config = {"database_key": None}

        return cls(**(kwargs | database_config | queue_config))

    async def _setup(self) -> None:
        await super()._setup()
        await self._create_table()
        await self._start_run()

    async def _destroy(self) -> None:
        await self._stop_run()
        await self._flush_queue()
        await super()._destroy()

    async def _create_table(self) -> None:
        operation = self.database_operation_factory.build_create()
        await self.execute_on_database(operation)

    async def _start_run(self) -> None:
        if self._run_task is None:
            self._run_task = create_task(self._run())

    async def _stop_run(self) -> None:
        if self._run_task is not None:
            task = self._run_task
            self._run_task = None

            task.cancel()
            with suppress(TimeoutError, CancelledError):
                await wait_for(task, 0.5)

    async def _flush_queue(self):
        while True:
            try:
                entry = self._queue.get_nowait()
            except QueueEmpty:
                break
            operation = self.database_operation_factory.build_mark_processed(entry.id_)
            await self.execute_on_database(operation)
            self._queue.task_done()

    async def _enqueue(self, message: BrokerMessage) -> None:
        operation = self.database_operation_factory.build_submit(message.topic, message.avro_bytes)
        await self.execute_on_database(operation)
        await self._notify_enqueued(message)

    # noinspection PyUnusedLocal
    async def _notify_enqueued(self, message: BrokerMessage) -> None:
        self._enqueued_event.set()

    async def _dequeue(self) -> BrokerMessage:
        while True:
            entry = await self._queue.get()
            try:
                # noinspection PyBroadException
                try:
                    message = entry.data
                except Exception as exc:
                    logger.warning(
                        f"There was a problem while trying to deserialize the entry with {entry.id_!r} id: {exc}"
                    )
                    operation = self.database_operation_factory.build_mark_processed(entry.id_)
                    await self.execute_on_database(operation)
                    continue

                operation = self.database_operation_factory.build_delete(entry.id_)
                await self.execute_on_database(operation)
                return message
            finally:
                self._queue.task_done()

    async def _run(self, max_wait: Optional[float] = 60.0) -> NoReturn:
        while self._run_task is not None:
            await self._wait_for_entries(max_wait)
            await self._dequeue_batch()

    async def _wait_for_entries(self, max_wait: Optional[float]) -> None:
        while True:
            if await self._get_count():
                return

            with suppress(TimeoutError):
                return await wait_for(self._wait_enqueued(), max_wait)

    async def _wait_enqueued(self) -> None:
        await self._enqueued_event.wait()
        self._enqueued_event.clear()

    async def _get_count(self) -> int:
        # noinspection PyTypeChecker
        operation = self.database_operation_factory.build_count(self.retry)
        row = await self.execute_on_database_and_fetch_one(operation)
        count = row[0]
        return count

    async def _dequeue_batch(self) -> None:
        async with self.database_pool.acquire() as client:
            rows = await self._dequeue_rows(client)

            if not len(rows):
                return

            entries = [_Entry(*row) for row in rows]

            ids = tuple(entry.id_ for entry in entries)
            operation = self.database_operation_factory.build_mark_processing(ids)
            await client.execute(operation)

        for entry in entries:
            await self._queue.put(entry)

    async def _dequeue_rows(self, client: DatabaseClient) -> list[Any]:
        operation = self.database_operation_factory.build_query(self._retry, self._records)
        await client.execute(operation)
        return [row async for row in client.fetch_all()]


class _Entry:
    def __init__(self, id_: int, data_bytes: bytes):
        self.id_ = id_
        self.data_bytes = data_bytes

    @cached_property
    def data(self) -> BrokerMessage:
        """Get the data.

        :return: A ``Model`` inherited instance.
        """
        return BrokerMessage.from_avro_bytes(self.data_bytes)

    def __lt__(self, other: Any) -> bool:
        # noinspection PyBroadException
        try:
            return isinstance(other, type(self)) and self.data < other.data
        except Exception:
            return False


class DatabaseBrokerQueueBuilder(Builder):
    """Database Broker Queue Builder class."""

    def with_config(self, config: Config):
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self.kwargs |= {"database_key": None}
        self.kwargs |= config.get_interface_by_name("broker").get("common", dict()).get("queue", dict())
        return super().with_config(config)


DatabaseBrokerQueue.set_builder(DatabaseBrokerQueueBuilder)



import logging
from abc import (
    ABC,
)

from ....collections import (
    BrokerQueue,
)

logger = logging.getLogger(__name__)


class BrokerPublisherQueue(BrokerQueue, ABC):
    """Broker Publisher Queue class."""



from .abc import (
    BrokerSubscriberDuplicateValidator,
)
from .database import (
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
)
from .memory import (
    InMemoryBrokerSubscriberDuplicateValidator,
)



from .impl import (
    FilteredBrokerSubscriber,
)
from .validators import (
    BrokerSubscriberDuplicateValidator,
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    BrokerSubscriberValidator,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
    InMemoryBrokerSubscriberDuplicateValidator,
)



from .abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)
from .database import (
    BrokerSubscriberQueueDatabaseOperationFactory,
    DatabaseBrokerSubscriberQueue,
    DatabaseBrokerSubscriberQueueBuilder,
)
from .memory import (
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from cached_property import (
    cached_property,
)

from ..requests import (
    Request,
    Response,
    ResponseException,
)


class HttpRequest(Request, ABC):
    """Http Request class."""

    @cached_property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        if "user" not in self.headers:
            return None
        return UUID(self.headers["user"])

    @property
    @abstractmethod
    def headers(self) -> dict[str, str]:
        """Get the headers of the request.

        :return: A dictionary in which keys are ``str`` instances and values are ``str`` instances.
        """

    @property
    @abstractmethod
    def content_type(self) -> str:
        """Get the content type.

        :return: A ``str`` value.
        """

    @abstractmethod
    async def url_params(self, type_: Optional[Union[type, str]] = None, **kwargs) -> Any:
        """Get the url params.

        :param type_: Optional ``type`` or ``str`` (classname) that defines the request content type.
        :param kwargs: Additional named arguments.
        :return: A dictionary instance.
        """

    @property
    @abstractmethod
    def has_url_params(self) -> bool:
        """Check if the request has url params.

        :return: ``True`` if it has url params or ``False`` otherwise.
        """

    @abstractmethod
    async def query_params(self, type_: Optional[Union[type, str]] = None, **kwargs) -> Any:
        """Get the query params.


        :param type_: Optional ``type`` or ``str`` (classname) that defines the request content type.
        :param kwargs: Additional named arguments.
        :return: A dictionary instance.
        """

    @property
    @abstractmethod
    def has_query_params(self) -> bool:
        # noinspection GrazieInspection
        """Check if the request has query params.

        :return: ``True`` if it has query params or ``False`` otherwise.
        """


class HttpResponse(Response, ABC):
    """Http Response class."""

    def __init__(self, *args, content_type: str = "application/json", **kwargs):
        super().__init__(*args, **kwargs)
        self._content_type = content_type

    @property
    def content_type(self) -> str:
        """Get the content type.

        :return: A ``str`` value.
        """
        return self._content_type

    @classmethod
    def from_response(cls, response: Optional[Response]) -> HttpResponse:
        """Build a new ``RestRequest`` from another response.

        :param response: The base response.
        :return: A ``RestResponse`` instance.
        """
        if isinstance(response, cls):
            return response

        if response is None:
            return cls()

        return cls(response._data, status=response.status)


class HttpResponseException(ResponseException):
    """Http Response Exception class."""



"""The networks core of the Minos Framework."""

__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.7.0.dev2"

from .brokers import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerClient,
    BrokerClientPool,
    BrokerDispatcher,
    BrokerHandler,
    BrokerHandlerService,
    BrokerMessage,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerMessageV1Status,
    BrokerMessageV1Strategy,
    BrokerPort,
    BrokerPublisher,
    BrokerPublisherBuilder,
    BrokerPublisherQueue,
    BrokerPublisherQueueDatabaseOperationFactory,
    BrokerQueue,
    BrokerQueueDatabaseOperationFactory,
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
    BrokerSubscriber,
    BrokerSubscriberBuilder,
    BrokerSubscriberDuplicateValidator,
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    BrokerSubscriberQueueDatabaseOperationFactory,
    BrokerSubscriberValidator,
    DatabaseBrokerPublisherQueue,
    DatabaseBrokerQueue,
    DatabaseBrokerQueueBuilder,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
    DatabaseBrokerSubscriberQueue,
    DatabaseBrokerSubscriberQueueBuilder,
    FilteredBrokerSubscriber,
    InMemoryBrokerPublisher,
    InMemoryBrokerPublisherQueue,
    InMemoryBrokerQueue,
    InMemoryBrokerSubscriber,
    InMemoryBrokerSubscriberBuilder,
    InMemoryBrokerSubscriberDuplicateValidator,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    QueuedBrokerPublisher,
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)
from .decorators import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    CheckDecorator,
    Checker,
    CheckerMeta,
    CheckerWrapper,
    EnrouteCollector,
    EnrouteDecorator,
    EnrouteDecoratorKind,
    EnrouteFactory,
    Handler,
    HandlerMeta,
    HandlerWrapper,
    HttpEnrouteDecorator,
    PeriodicEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
    enroute,
)
from .discovery import (
    DiscoveryClient,
    DiscoveryConnector,
    InMemoryDiscoveryClient,
)
from .exceptions import (
    MinosActionNotFoundException,
    MinosDiscoveryConnectorException,
    MinosHandlerException,
    MinosHandlerNotFoundEnoughEntriesException,
    MinosInvalidDiscoveryClient,
    MinosMultipleEnrouteDecoratorKindsException,
    MinosNetworkException,
    MinosRedefinedEnrouteDecoratorException,
    NotHasContentException,
    NotHasParamsException,
    NotSatisfiedCheckerException,
    RequestException,
)
from .http import (
    HttpAdapter,
    HttpConnector,
    HttpPort,
    HttpRequest,
    HttpResponse,
    HttpResponseException,
    RestService,
)
from .requests import (
    REQUEST_USER_CONTEXT_VAR,
    InMemoryRequest,
    Request,
    Response,
    ResponseException,
    WrappedRequest,
)
from .routers import (
    BrokerRouter,
    HttpRouter,
    PeriodicRouter,
    RestHttpRouter,
    Router,
)
from .scheduling import (
    CronTab,
    PeriodicPort,
    PeriodicTask,
    PeriodicTaskScheduler,
    PeriodicTaskSchedulerService,
    ScheduledRequest,
    ScheduledRequestContent,
    ScheduledResponseException,
)
from .specs import (
    AsyncAPIService,
    OpenAPIService,
)
from .system import (
    SystemService,
)
from .utils import (
    consume_queue,
    get_host_ip,
    get_host_name,
    get_ip,
)



from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Iterable,
)

from minos.common import (
    DatabaseOperation,
)

from .....collections import (
    BrokerQueueDatabaseOperationFactory,
)


# noinspection SqlNoDataSourceInspection,SqlResolve,PyTypeChecker,PyArgumentList
class BrokerSubscriberQueueDatabaseOperationFactory(BrokerQueueDatabaseOperationFactory, ABC):
    """Broker Subscriber Queue Database Operation Factory class."""

    @abstractmethod
    def build_count(
        self,
        retry: int,
        topics: Iterable[str] = tuple(),
        *args,
        **kwargs,
    ) -> DatabaseOperation:
        """Build the "count not processed" query.

        :return:
        """

    @abstractmethod
    def build_query(
        self,
        retry: int,
        records: int,
        topics: Iterable[str] = tuple(),
        *args,
        **kwargs,
    ) -> DatabaseOperation:
        """Build the "select not processed" query.

        :return: A ``SQL`` instance.
        """



from __future__ import (
    annotations,
)

import logging
from inspect import (
    isclass,
)
from itertools import (
    chain,
)
from operator import (
    itemgetter,
)
from typing import (
    Any,
    Optional,
)

from minos.common import (
    Config,
    Injectable,
    SetupMixin,
)

from ..decorators import (
    EnrouteCollector,
)
from ..exceptions import (
    MinosInvalidDiscoveryClient,
)
from ..utils import (
    get_host_ip,
)
from .clients import (
    DiscoveryClient,
)

logger = logging.getLogger(__name__)


@Injectable("discovery")
class DiscoveryConnector(SetupMixin):
    """Discovery Connector class."""

    def __init__(
        self,
        client: DiscoveryClient,
        name: str,
        endpoints: list[dict[str, Any]],
        host: str,
        port: Optional[int] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if port is None:
            port = 8080

        self.client = client

        self.name = name
        self.host = host
        self.port = port
        self.endpoints = endpoints

    @classmethod
    def _from_config(cls, *args, config: Config, **kwargs) -> DiscoveryConnector:
        client = cls._client_from_config(config)
        port = cls._port_from_config(config)
        name = config.get_name()
        host = get_host_ip()
        endpoints = cls._endpoints_from_config(config)

        return cls(client, name, endpoints, host, port, *args, **kwargs)

    @classmethod
    def _client_from_config(cls, config: Config) -> DiscoveryClient:
        discovery_config = config.get_discovery()

        client_cls = cls._client_cls_from_config(discovery_config)
        return client_cls.from_config(config)

    @staticmethod
    def _client_cls_from_config(discovery_config: dict[str, Any]) -> type[DiscoveryClient]:
        client_cls = discovery_config["client"]
        if not isclass(client_cls) or not issubclass(client_cls, DiscoveryClient):
            raise MinosInvalidDiscoveryClient(f"{client_cls!r} not supported.")
        return client_cls

    @staticmethod
    def _port_from_config(config: Config) -> Optional[int]:
        http_config = config.get_interface_by_name("http")
        connector_config = http_config["connector"]
        port = connector_config.get("port")
        return port

    @staticmethod
    def _endpoints_from_config(config: Config) -> list[dict[str, Any]]:
        endpoints = list()
        for name in config.get_services():
            decorators = EnrouteCollector(name, config).get_rest_command_query()
            endpoints += [
                {"url": decorator.url, "method": decorator.method} | decorator.kwargs
                for decorator in set(chain(*decorators.values()))
            ]

        endpoints.sort(key=itemgetter("url", "method"))
        return endpoints

    async def _setup(self) -> None:
        await self.subscribe()

    async def subscribe(self) -> None:
        """Send a subscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery subscription...")
        await self.client.subscribe(self.host, self.port, self.name, self.endpoints)

    async def _destroy(self) -> None:
        await self.unsubscribe()

    async def unsubscribe(self) -> None:
        """Send an unsubscribe operation to the discovery.

        :return: This method does not return anything.
        """
        logger.info("Performing discovery unsubscription...")
        await self.client.unsubscribe(self.name)



from collections.abc import (
    Iterable,
)

from minos.common import (
    DatabaseOperation,
)
from minos.common.testing import (
    MockedDatabaseClient,
    MockedDatabaseOperation,
)

from ....brokers import (
    BrokerSubscriberQueueDatabaseOperationFactory,
)
from ..collections import (
    MockedBrokerQueueDatabaseOperationFactory,
)


class MockedBrokerSubscriberQueueDatabaseOperationFactory(
    BrokerSubscriberQueueDatabaseOperationFactory, MockedBrokerQueueDatabaseOperationFactory
):
    """For testing purposes"""

    def build_count(self, retry: int, topics: Iterable[str] = tuple(), *args, **kwargs) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("count_not_processed")

    def build_query(
        self, retry: int, records: int, topics: Iterable[str] = tuple(), *args, **kwargs
    ) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("select_not_processed")


MockedDatabaseClient.set_factory(
    BrokerSubscriberQueueDatabaseOperationFactory, MockedBrokerSubscriberQueueDatabaseOperationFactory
)



from collections.abc import (
    Iterable,
)

from minos.common import (
    DatabaseOperation,
)
from minos.common.testing import (
    MockedDatabaseOperation,
)

from ....brokers import (
    BrokerQueueDatabaseOperationFactory,
)


class MockedBrokerQueueDatabaseOperationFactory(BrokerQueueDatabaseOperationFactory):
    """For testing purposes."""

    def build_create(self) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_queue_table")

    def build_mark_processed(self, id_: int) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("update_not_processed")

    def build_delete(self, id_: int) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("delete_processed")

    def build_mark_processing(self, ids: Iterable[int]) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("mark_processing")

    def build_count(self, retry: int, *args, **kwargs) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("count_not_processed")

    def build_submit(self, topic: str, data: bytes) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("insert")

    def build_query(self, retry: int, records: int, *args, **kwargs) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("select_not_processed")



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    AsyncIterator,
    Iterable,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from minos.common import (
    BuildableMixin,
    Builder,
    Config,
    Injectable,
    MinosConfigException,
)

from ..messages import (
    BrokerMessage,
)

if TYPE_CHECKING:
    from .filtered import (
        BrokerSubscriberValidator,
        FilteredBrokerSubscriber,
    )
    from .queued import (
        BrokerSubscriberQueue,
        BrokerSubscriberQueueBuilder,
        QueuedBrokerSubscriber,
    )


logger = logging.getLogger(__name__)


class BrokerSubscriber(ABC, BuildableMixin):
    """Broker Subscriber class."""

    def __init__(self, topics: Iterable[str], **kwargs):
        super().__init__(**kwargs)
        self._topics = set(topics)

    @property
    def topics(self) -> set[str]:
        """Topics getter.

        :return: A list of string values.
        """
        return self._topics

    def __aiter__(self) -> AsyncIterator[BrokerMessage]:
        return self

    async def __anext__(self) -> BrokerMessage:
        if self.already_destroyed:
            raise StopAsyncIteration
        return await self.receive()

    async def receive(self) -> BrokerMessage:
        """Receive a new message.

        :return: A ``BrokerMessage`` instance.
        """
        message = await self._receive()
        logger.debug(f"Receiving {message!r} message...")
        return message

    @abstractmethod
    async def _receive(self) -> BrokerMessage:
        raise NotImplementedError


BrokerSubscriberCls = TypeVar("BrokerSubscriberCls", bound=BrokerSubscriber)


@Injectable("broker_subscriber_builder")
class BrokerSubscriberBuilder(Builder[BrokerSubscriberCls], Generic[BrokerSubscriberCls]):
    """Broker Subscriber Builder class."""

    def __init__(
        self,
        *args,
        validator_builder: Optional[Builder] = None,
        queue_builder: Optional[BrokerSubscriberQueueBuilder] = None,
        filtered_cls: Optional[type[FilteredBrokerSubscriber]] = None,
        queued_cls: Optional[type[QueuedBrokerSubscriber]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if filtered_cls is None:
            from .filtered import (
                FilteredBrokerSubscriber,
            )

            filtered_cls = FilteredBrokerSubscriber

        if queued_cls is None:
            from .queued import (
                QueuedBrokerSubscriber,
            )

            queued_cls = QueuedBrokerSubscriber

        self.validator_builder = validator_builder
        self.queue_builder = queue_builder

        self.filtered_cls = filtered_cls
        self.queued_cls = queued_cls

    def with_filtered_cls(self, filtered_cls: type[FilteredBrokerSubscriber]):
        """Set the filtered class.

        :param filtered_cls: A subclass of ``FilteredBrokerSubscriber``.
        :return: This method return the builder instance.
        """
        self.filtered_cls = filtered_cls

        return self

    def with_queued_cls(self, queued_cls: type[QueuedBrokerSubscriber]):
        """Set the queued class.

        :param queued_cls: A subclass of ``QueuedBrokerSubscriber``.
        :return: This method return the builder instance.
        """
        self.queued_cls = queued_cls

        return self

    def with_config(self, config: Config):
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self._with_builders_from_config(config)

        if self.validator_builder is not None:
            self.validator_builder.with_config(config)
        if self.queue_builder is not None:
            self.queue_builder.with_config(config)
        return super().with_config(config)

    def _with_builders_from_config(self, config):
        try:
            broker_config = config.get_interface_by_name("broker")
        except MinosConfigException:
            return

        broker_subscriber_config = broker_config["subscriber"]

        if "validator" in broker_subscriber_config:
            self.with_validator(broker_subscriber_config["validator"])

        if "queue" in broker_subscriber_config:
            self.with_queue(broker_subscriber_config["queue"])

    def with_validator(
        self,
        validator: Union[type[BrokerSubscriberValidator], Builder[BrokerSubscriberValidator]],
    ):
        """Set the duplicate detector.

        :param validator: The duplicate detector to be set.
        :return: This method return the builder instance.
        """
        if not isinstance(validator, Builder):
            validator = validator.get_builder()
        self.validator_builder = validator.copy()
        return self

    def with_queue(self, queue: Union[type[BrokerSubscriberQueue], BrokerSubscriberQueueBuilder]):
        """Set the queue builder.

        :param queue: The queue to be set.
        :return: This method return the builder instance.
        """
        if not isinstance(queue, Builder):
            queue = queue.get_builder()
        self.queue_builder = queue.copy()
        return self

    def with_kwargs(self, kwargs: dict[str, Any]):
        """Set kwargs.

        :param kwargs: The kwargs to be set.
        :return: This method return the builder instance.
        """
        if self.validator_builder is not None:
            self.validator_builder.with_kwargs(kwargs)

        if self.queue_builder is not None:
            self.queue_builder.with_kwargs(kwargs)

        return super().with_kwargs(kwargs)

    def with_group_id(self, group_id: Optional[str]):
        """Set group_id.

        :param group_id: The group_id to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["group_id"] = group_id

        return self

    def with_remove_topics_on_destroy(self, remove_topics_on_destroy: bool):
        """Set remove_topics_on_destroy.

        :param remove_topics_on_destroy: The remove_topics_on_destroy flag to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["remove_topics_on_destroy"] = remove_topics_on_destroy

        return self

    def with_topics(self, topics: Iterable[str]):
        """Set topics.

        :param topics: The topics to be set.
        :return: This method return the builder instance.
        """
        topics = set(topics)
        self.kwargs["topics"] = set(topics)

        if self.queue_builder is not None:
            self.queue_builder.with_topics(topics)

        return self

    def build(self) -> BrokerSubscriber:
        """Build the instance.

        :return: A ``QueuedBrokerSubscriber`` instance.
        """
        impl = super().build()

        if self.validator_builder is not None:
            validator = self.validator_builder.build()
            impl = self.filtered_cls(impl=impl, validator=validator, **self.kwargs)

        if self.queue_builder is not None:
            queue = self.queue_builder.build()
            impl = self.queued_cls(impl=impl, queue=queue, **self.kwargs)

        return impl


BrokerSubscriber.set_builder(BrokerSubscriberBuilder)



from __future__ import (
    annotations,
)

from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    Builder,
    Config,
    DatabaseMixin,
    IntegrityException,
)

from ..abc import (
    BrokerSubscriberDuplicateValidator,
)
from .factories import (
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
)


class DatabaseBrokerSubscriberDuplicateValidator(
    BrokerSubscriberDuplicateValidator,
    DatabaseMixin[BrokerSubscriberDuplicateValidatorDatabaseOperationFactory],
):
    """Database Broker Subscriber Duplicate Detector class."""

    def __init__(self, *args, database_key: Optional[tuple[str]] = None, **kwargs):
        if database_key is None:
            database_key = ("broker",)
        super().__init__(*args, database_key=database_key, **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        await self._create_table()

    async def _create_table(self) -> None:
        operation = self.database_operation_factory.build_create()
        await self.execute_on_database(operation)

    async def _is_unique(self, topic: str, uuid: UUID) -> bool:
        operation = self.database_operation_factory.build_submit(topic, uuid)
        try:
            await self.execute_on_database(operation)
            return True
        except IntegrityException:
            return False


class DatabaseBrokerSubscriberDuplicateValidatorBuilder(Builder[DatabaseBrokerSubscriberDuplicateValidator]):
    """Database Broker Subscriber Duplicate Detector Builder class."""

    def with_config(self, config: Config):
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self.kwargs |= {"database_key": None}
        return super().with_config(config)


DatabaseBrokerSubscriberDuplicateValidator.set_builder(DatabaseBrokerSubscriberDuplicateValidatorBuilder)



from .abc import (
    DiscoveryClient,
)


class InMemoryDiscoveryClient(DiscoveryClient):
    """In Memory Discovery Client class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_subscribed = False

    @property
    def is_subscribed(self) -> bool:
        """Check if the client is subscribed or not.

        :return:
        """
        return self._is_subscribed

    async def subscribe(self, *args, **kwargs) -> None:
        """Subscribe to the discovery.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        self._is_subscribed = True

    async def unsubscribe(self, *args, **kwargs) -> None:
        """Unsubscribe from the discovery.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        self._is_subscribed = False



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)

from minos.common import (
    BuildableMixin,
)

from ....messages import (
    BrokerMessage,
)


class BrokerSubscriberValidator(BuildableMixin, ABC):
    """Broker Subscriber Duplicate Detector class."""

    async def is_valid(self, message: BrokerMessage) -> bool:
        """Check if the given message is valid.

        :param message: The message to be checked.
        :return: ``True`` if it is valid or ``False`` otherwise.
        """
        return await self._is_valid(message)

    @abstractmethod
    async def _is_valid(self, message: BrokerMessage) -> bool:
        raise NotImplementedError



from __future__ import (
    annotations,
)

import logging
from collections.abc import (
    Callable,
)
from typing import (
    Any,
)

from minos.common import (
    Config,
    SetupMixin,
    import_module,
)

from ..decorators import (
    HttpEnrouteDecorator,
)
from ..exceptions import (
    MinosRedefinedEnrouteDecoratorException,
)
from ..routers import (
    HttpRouter,
)

logger = logging.getLogger(__name__)


class HttpAdapter(SetupMixin):
    """Http Adapter class."""

    def __init__(self, routers: list[HttpRouter], **kwargs):
        super().__init__(**kwargs)
        self._routers = routers

    @classmethod
    def _from_config(cls, *args, config: Config, **kwargs) -> HttpAdapter:
        routers = cls._routers_from_config(config)
        return cls(routers=routers, **kwargs)

    @staticmethod
    def _routers_from_config(config: Config, **kwargs) -> list[HttpRouter]:
        classes = config.get_routers()
        classes = tuple((class_ if not isinstance(class_, str) else import_module(class_)) for class_ in classes)
        classes = filter(lambda router: issubclass(router, HttpRouter), classes)
        routers = [router.from_config(config) for router in classes]
        return routers

    @property
    def routes(self) -> dict[HttpEnrouteDecorator, Callable]:
        """Get routes.

        :return: A ``dict`` with ``HttpEnrouteDecorator`` and ``Callable`` as values.
        """
        routes = dict()
        for router in self._routers:
            for decorator, callback in router.routes.items():
                if decorator in routes:
                    raise MinosRedefinedEnrouteDecoratorException(f"{decorator!r} can be used only once.")
                routes[decorator] = callback
        return routes

    @property
    def routers(self) -> list[HttpRouter]:
        """Get the routers.

        :return: A ``list`` of ``Router`` instances.
        """
        return self._routers

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self._routers == other._routers



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from contextvars import (
    ContextVar,
)
from typing import (
    Any,
    Final,
    Optional,
)
from uuid import (
    UUID,
)

from ..exceptions import (
    MinosException,
    NotHasContentException,
    NotHasParamsException,
)

REQUEST_USER_CONTEXT_VAR: Final[ContextVar[Optional[UUID]]] = ContextVar("user", default=None)


class Request(ABC):
    """Request interface."""

    @property
    @abstractmethod
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        raise NotImplementedError

    async def content(self, **kwargs) -> Any:
        """Get the request content.

        :param kwargs: Additional named arguments.
        :return: The request content.
        """
        if not self.has_content:
            raise NotHasContentException(f"{self!r} has not content.")
        return await self._content(**kwargs)

    @property
    @abstractmethod
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        raise NotImplementedError

    async def _content(self, **kwargs) -> Any:
        raise RuntimeError(
            f"{type(self).__name__}._content must be implemented if {type(self).__name__}.has_content returns {True!r}."
        )

    async def params(self, **kwargs) -> dict[str, Any]:
        """Get the request params.

        :param kwargs: Additional named arguments.
        :return: The request params.
        """
        if not self.has_params:
            raise NotHasParamsException(f"{self!r} has not params.")
        return await self._params(**kwargs)

    @property
    @abstractmethod
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        raise NotImplementedError

    async def _params(self, **kwargs) -> dict[str, Any]:
        raise RuntimeError(
            f"{type(self).__name__}._params must be implemented if '{type(self).__name__}.has_params returns {True!r}."
        )

    @abstractmethod
    def __eq__(self, other: Request) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


sentinel = object()


class Response:
    """Response definition."""

    __slots__ = "_data", "_status"

    def __init__(self, data: Any = sentinel, *, status: int = 200):
        if not isinstance(status, int):
            raise ValueError(f"The status argument must be integer. Obtained: {status}")
        self._data = data
        self._status = status

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Any:
        """Response content.

        :param kwargs: Additional named arguments.
        :return: A list of items.
        """
        if not self.has_content:
            return None
        return self._data

    @property
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        return self._data is not sentinel

    @property
    def status(self) -> int:
        """The status code of the response.

        :return: An ``int`` value.
        """
        return self._status

    def __eq__(self, other: Response) -> bool:
        return type(self) == type(other) and self._data == other._data and self._status == other._status

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._data!r})"

    def __hash__(self):
        return hash(self._data)


class ResponseException(MinosException):
    """Response Exception class."""

    def __init__(self, *args, status: int = 400):
        super().__init__(*args)
        self._status = status

    @property
    def status(self) -> int:
        """The status code of the response.

        :return: An ``int`` value.
        """
        return self._status



from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Iterable,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


# noinspection SqlResolve,SqlNoDataSourceInspection,SqlNoDataSourceInspection,SqlResolve
class BrokerQueueDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Broker Queue Database Operation Factory class."""

    @abstractmethod
    def build_create(self) -> DatabaseOperation:
        """Build the "create table" query.

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_mark_processed(self, id_: int) -> DatabaseOperation:
        """Build the "update not processed" query.

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_delete(self, id_: int) -> DatabaseOperation:
        """Build the "delete processed" query.

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_mark_processing(self, ids: Iterable[int]) -> DatabaseOperation:
        """

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_count(self, retry: int, *args, **kwargs) -> DatabaseOperation:
        """Build the "count not processed" query.

        :return:
        """

    @abstractmethod
    def build_submit(self, topic: str, data: bytes) -> DatabaseOperation:
        """Build the "insert" query.

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_query(self, retry: int, records: int, *args, **kwargs) -> DatabaseOperation:
        """Build the "select not processed" query.

        :return: A ``SQL`` instance.
        """



from .abc import (
    BrokerPublisher,
    BrokerPublisherBuilder,
)
from .memory import (
    InMemoryBrokerPublisher,
)
from .queued import (
    BrokerPublisherQueue,
    BrokerPublisherQueueDatabaseOperationFactory,
    DatabaseBrokerPublisherQueue,
    InMemoryBrokerPublisherQueue,
    QueuedBrokerPublisher,
)



from __future__ import (
    annotations,
)

import logging
import warnings
from typing import (
    Optional,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Inject,
    NotProvidedException,
    Port,
)

from .connectors import (
    HttpConnector,
)

logger = logging.getLogger(__name__)


class HttpPort(Port):
    """Http Port class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_kwargs = kwargs

    async def _start(self) -> None:
        await self.connector.setup()
        await self.connector.start()

    async def _stop(self, err: Exception = None) -> None:
        await self.connector.stop()
        await self.connector.destroy()

    @cached_property
    def connector(self) -> HttpConnector:
        """Get the port connector.

        :return: A ``HttpConnector`` instance.
        """
        return self._get_connector(**self._init_kwargs)

    @staticmethod
    @Inject()
    def _get_connector(
        connector: Optional[HttpConnector] = None,
        http_connector: Optional[HttpConnector] = None,
        **kwargs,
    ) -> HttpConnector:
        if connector is None:
            connector = http_connector
        if connector is None:
            raise NotProvidedException(f"A {HttpConnector!r} must be provided.")
        return connector


class RestService(HttpPort):
    """Rest Service class."""

    def __init__(self, **kwargs):
        warnings.warn(f"{RestService!r} has been deprecated. Use {HttpPort} instead.", DeprecationWarning)
        super().__init__(**kwargs)



from .clients import (
    BrokerClient,
)
from .collections import (
    BrokerQueue,
    BrokerQueueDatabaseOperationFactory,
    DatabaseBrokerQueue,
    DatabaseBrokerQueueBuilder,
    InMemoryBrokerQueue,
)
from .dispatchers import (
    BrokerDispatcher,
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
)
from .handlers import (
    BrokerHandler,
    BrokerHandlerService,
    BrokerPort,
)
from .messages import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerMessage,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerMessageV1Status,
    BrokerMessageV1Strategy,
)
from .pools import (
    BrokerClientPool,
)
from .publishers import (
    BrokerPublisher,
    BrokerPublisherBuilder,
    BrokerPublisherQueue,
    BrokerPublisherQueueDatabaseOperationFactory,
    DatabaseBrokerPublisherQueue,
    InMemoryBrokerPublisher,
    InMemoryBrokerPublisherQueue,
    QueuedBrokerPublisher,
)
from .subscribers import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
    BrokerSubscriberDuplicateValidator,
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    BrokerSubscriberQueueDatabaseOperationFactory,
    BrokerSubscriberValidator,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
    DatabaseBrokerSubscriberQueue,
    DatabaseBrokerSubscriberQueueBuilder,
    FilteredBrokerSubscriber,
    InMemoryBrokerSubscriber,
    InMemoryBrokerSubscriberBuilder,
    InMemoryBrokerSubscriberDuplicateValidator,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)



from .abc import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)
from .filtered import (
    BrokerSubscriberDuplicateValidator,
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    BrokerSubscriberValidator,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
    FilteredBrokerSubscriber,
    InMemoryBrokerSubscriberDuplicateValidator,
)
from .memory import (
    InMemoryBrokerSubscriber,
    InMemoryBrokerSubscriberBuilder,
)
from .queued import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    BrokerSubscriberQueueDatabaseOperationFactory,
    DatabaseBrokerSubscriberQueue,
    DatabaseBrokerSubscriberQueueBuilder,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)



from .abc import (
    BrokerPublisherQueue,
)
from .database import (
    BrokerPublisherQueueDatabaseOperationFactory,
    DatabaseBrokerPublisherQueue,
)
from .memory import (
    InMemoryBrokerPublisherQueue,
)



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    Queue,
    create_task,
    gather,
)
from collections.abc import (
    Iterable,
)
from typing import (
    NoReturn,
    Optional,
)

from minos.common import (
    Config,
    Inject,
    NotProvidedException,
    SetupMixin,
)

from ..dispatchers import (
    BrokerDispatcher,
)
from ..subscribers import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)

logger = logging.getLogger(__name__)


class BrokerHandler(SetupMixin):
    """Broker Handler class."""

    def __init__(
        self, dispatcher: BrokerDispatcher, subscriber: BrokerSubscriber, concurrency: int = 5, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._dispatcher = dispatcher
        self._subscriber = subscriber

        self._queue = Queue(maxsize=1)
        self._consumers = list()
        self._concurrency = concurrency

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> BrokerHandler:
        dispatcher = cls._get_dispatcher(config, **kwargs)
        subscriber = cls._get_subscriber(config, topics=set(dispatcher.actions.keys()), **kwargs)

        return cls(dispatcher, subscriber, **kwargs)

    @staticmethod
    @Inject()
    def _get_dispatcher(
        config: Config,
        dispatcher: Optional[BrokerDispatcher] = None,
        broker_dispatcher: Optional[BrokerDispatcher] = None,
        **kwargs,
    ) -> BrokerDispatcher:
        if dispatcher is None:
            dispatcher = broker_dispatcher
        if dispatcher is None:
            dispatcher = BrokerDispatcher.from_config(config, **kwargs)
        return dispatcher

    @staticmethod
    @Inject()
    def _get_subscriber(
        config: Config,
        topics: Iterable[str],
        subscriber: Optional[BrokerSubscriber] = None,
        broker_subscriber: Optional[BrokerSubscriber] = None,
        subscriber_builder: Optional[BrokerSubscriberBuilder] = None,
        broker_subscriber_builder: Optional[BrokerSubscriberBuilder] = None,
        **kwargs,
    ) -> BrokerSubscriber:
        if not isinstance(subscriber, BrokerSubscriber):
            subscriber = broker_subscriber

        if not isinstance(subscriber, BrokerSubscriber):
            if not isinstance(subscriber_builder, BrokerSubscriberBuilder):
                subscriber_builder = broker_subscriber_builder

            if isinstance(subscriber_builder, BrokerSubscriberBuilder):
                subscriber = (
                    subscriber_builder.copy().with_config(config).with_topics(topics).with_kwargs(kwargs).build()
                )

        if not isinstance(subscriber, BrokerSubscriber):
            raise NotProvidedException(f"A {BrokerSubscriber!r} or {BrokerSubscriberBuilder!r} must be provided.")

        return subscriber

    async def _setup(self) -> None:
        await super()._setup()
        await self._dispatcher.setup()

        await self._create_consumers()

        await self._subscriber.setup()

    async def _destroy(self) -> None:
        await self._subscriber.destroy()

        await self._destroy_consumers()

        await self._dispatcher.destroy()
        await super()._destroy()

    async def run(self) -> NoReturn:
        """Run the message handling process.

        :return: This method does not return anything.
        """
        async for message in self._subscriber:
            await self._queue.put(message)

    async def _create_consumers(self):
        while len(self._consumers) < self._concurrency:
            self._consumers.append(create_task(self._consume()))

    async def _destroy_consumers(self):
        await self._queue.join()
        for consumer in self._consumers:
            consumer.cancel()
        await gather(*self._consumers, return_exceptions=True)
        self._consumers = list()

    async def _consume(self) -> None:
        while True:
            await self._consume_one()

    async def _consume_one(self) -> None:
        message = await self._queue.get()
        try:
            try:
                await self._dispatcher.dispatch(message)
            except Exception as exc:
                logger.warning(f"An exception was raised: {exc!r}")
        finally:
            self._queue.task_done()



from .collections import (
    MockedBrokerQueueDatabaseOperationFactory,
)
from .publishers import (
    MockedBrokerPublisherQueueDatabaseOperationFactory,
)
from .subscribers import (
    MockedBrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    MockedBrokerSubscriberQueueDatabaseOperationFactory,
)



from __future__ import (
    annotations,
)

import logging

from ..messages import (
    BrokerMessage,
)
from .abc import (
    BrokerPublisher,
)

logger = logging.getLogger(__name__)


class InMemoryBrokerPublisher(BrokerPublisher):
    """In Memory Broker Publisher class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._messages = list()

    async def _send(self, message: BrokerMessage) -> None:
        """Send a message.

        :param message: The message to be sent.
        :return: This method does not return anything.
        """
        self._messages.append(message)

    @property
    def messages(self) -> list[BrokerMessage]:
        """The sequence of sent messages.

        :return: A list of ``BrokerMessage`` entries.
        """
        return self._messages



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)

from minos.common import (
    Config,
    SetupMixin,
)

logger = logging.getLogger(__name__)


class DiscoveryClient(ABC, SetupMixin):
    """Discovery Client class."""

    def __init__(self, host: str, port: int, **kwargs):
        super().__init__(**kwargs)
        self.host = host
        self.port = port

    @property
    def route(self) -> str:
        """Get the full http route to the discovery.

        :return: An ``str`` value.
        """
        # noinspection HttpUrlsUsage
        return f"http://{self.host}:{self.port}"

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> DiscoveryClient:
        discovery_config = config.get_discovery()

        client_host = discovery_config.get("host")
        client_port = discovery_config.get("port")
        return cls(host=client_host, port=client_port, **kwargs)

    @abstractmethod
    async def subscribe(
        self, host: str, port: int, name: str, endpoints: list[dict[str, str]], *args, **kwargs
    ) -> None:
        """Subscribe to the discovery.

        :param host: The ip of the microservice to be subscribed.
        :param port: The port of the microservice to be subscribed.
        :param name: The name of the microservice to be subscribed.
        :param endpoints: List of endpoints exposed by the microservice.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """

    @abstractmethod
    async def unsubscribe(self, name: str, *args, **kwargs) -> None:
        """Unsubscribe from the discovery.

        :param name: The name of the microservice to be unsubscribed.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """



from collections.abc import (
    Iterable,
)
from uuid import (
    UUID,
)

from .abc import (
    BrokerSubscriberDuplicateValidator,
)


class InMemoryBrokerSubscriberDuplicateValidator(BrokerSubscriberDuplicateValidator):
    """In Memory Broker Subscriber Duplicate Detector class."""

    def __init__(self, seen: Iterable[tuple[str, UUID]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if seen is None:
            seen = set()
        self._seen = set(seen)

    @property
    def seen(self) -> set[tuple[str, UUID]]:
        """Get the seen pairs.

        :return: A ``set`` of ``tuple`` instances in which the first value is a ``str`` and the second an ``UUID``.
        """
        return self._seen

    async def _is_unique(self, topic: str, uuid: UUID) -> bool:
        if (topic, uuid) not in self._seen:
            self._seen.add((topic, uuid))
            return True
        return False



from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    DeclarativeModel,
)

from ..requests import (
    Request,
    ResponseException,
)


class ScheduledRequest(Request):
    """Scheduling Request class."""

    def __init__(self, scheduled_at: datetime, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scheduled_at = scheduled_at

    @property
    def user(self) -> Optional[UUID]:
        """The user of the request.

        :return: Always return ``None`` as scheduled request are launched by the system.
        """
        return None

    @property
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        return True

    async def _content(self, **kwargs) -> ScheduledRequestContent:
        return self._content_value

    @property
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        return False

    def __eq__(self, other: Request) -> bool:
        return isinstance(other, type(self)) and self._content_value == other._content_value

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._content_value!r})"

    @cached_property
    def _content_value(self) -> ScheduledRequestContent:
        return ScheduledRequestContent(self._scheduled_at)


class ScheduledRequestContent(DeclarativeModel):
    """Scheduling Request Content class."""

    scheduled_at: datetime


class ScheduledResponseException(ResponseException):
    """Scheduled Response Exception class."""



from __future__ import (
    annotations,
)

import logging

from ....collections import (
    InMemoryBrokerQueue,
)
from .abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)

logger = logging.getLogger(__name__)


class InMemoryBrokerSubscriberQueue(InMemoryBrokerQueue, BrokerSubscriberQueue):
    """In Memory Broker Subscriber Queue class."""


class InMemoryBrokerSubscriberQueueBuilder(BrokerSubscriberQueueBuilder[InMemoryBrokerSubscriberQueue]):
    """In Memory Broker Subscriber Queue Builder class."""


InMemoryBrokerSubscriberQueue.set_builder(InMemoryBrokerSubscriberQueueBuilder)



from __future__ import (
    annotations,
)

from contextlib import (
    suppress,
)
from typing import (
    Optional,
    Protocol,
)

from minos.common import (
    get_internal_modules,
)

from .definitions import (
    BrokerCommandEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    EnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestQueryEnrouteDecorator,
)


class SubEnroute(Protocol):
    """Sub Enroute protocol class."""

    command: Optional[EnrouteDecorator]
    query: Optional[EnrouteDecorator]
    event: Optional[EnrouteDecorator]


class BrokerEnroute:
    """Broker Enroute class"""

    command = BrokerCommandEnrouteDecorator
    query = BrokerQueryEnrouteDecorator
    event = BrokerEventEnrouteDecorator


class RestEnroute:
    """Rest Enroute class"""

    command = RestCommandEnrouteDecorator
    query = RestQueryEnrouteDecorator


class PeriodicEnroute:
    """Periodic Enroute class."""

    event = PeriodicEventEnrouteDecorator


class EnrouteType(type):
    """Enroute type."""

    def __getattr__(cls, item: str) -> SubEnroute:
        for module in get_internal_modules():
            with suppress(AttributeError):
                # noinspection PyProtectedMember
                module._register_enroute()
        with suppress(AttributeError):
            return object.__getattribute__(cls, item)
        raise AttributeError(f"{item} not in enroute.")


class Enroute(metaclass=EnrouteType):
    """Enroute decorator main class"""

    broker = BrokerEnroute
    rest = RestEnroute
    periodic = PeriodicEnroute

    @classmethod
    def _register_sub_enroute(cls, name: str, sub_enroute) -> None:
        """Register a new sub enroute.

        :param name: The name to be registered.
        :param sub_enroute: Teh sub enroute to be registered.
        :return: This method does not return anything.
        """
        setattr(cls, name, sub_enroute)

    @classmethod
    def _unregister_sub_enroute(cls, name: str) -> None:
        """Unregister a new sub enroute.

        :param name: The name to be unregistered.
        :return: This method does not return anything.
        """
        delattr(cls, name)


enroute = Enroute



from minos.common.testing import (
    MockedDatabaseClient,
)
from minos.networks import (
    BrokerPublisherQueueDatabaseOperationFactory,
)

from ..collections import (
    MockedBrokerQueueDatabaseOperationFactory,
)


class MockedBrokerPublisherQueueDatabaseOperationFactory(
    BrokerPublisherQueueDatabaseOperationFactory, MockedBrokerQueueDatabaseOperationFactory
):
    """For testing purposes"""


MockedDatabaseClient.set_factory(
    BrokerPublisherQueueDatabaseOperationFactory, MockedBrokerPublisherQueueDatabaseOperationFactory
)



from __future__ import (
    annotations,
)

import logging
import traceback
from abc import (
    ABC,
    abstractmethod,
)
from asyncio import (
    Semaphore,
)
from collections.abc import (
    Callable,
)
from functools import (
    wraps,
)
from inspect import (
    isawaitable,
)
from typing import (
    Awaitable,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from minos.common import (
    Config,
    Injectable,
    SetupMixin,
)

from ..decorators import (
    HttpEnrouteDecorator,
)
from ..requests import (
    REQUEST_USER_CONTEXT_VAR,
    Request,
    Response,
    ResponseException,
)
from .adapters import (
    HttpAdapter,
)

_Callback = Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]

RawRequest = TypeVar("RawRequest")
RawResponse = TypeVar("RawResponse")

logger = logging.getLogger(__name__)


@Injectable("http_connector")
class HttpConnector(ABC, SetupMixin, Generic[RawRequest, RawResponse]):
    """Http Application base class."""

    def __init__(
        self,
        adapter: HttpAdapter,
        host: Optional[str] = None,
        port: Optional[int] = None,
        max_connections: int = 5,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if host is None:
            host = "0.0.0.0"
        if port is None:
            port = 8080

        self._adapter = adapter
        self._host = host
        self._port = port

        self._semaphore = Semaphore(max_connections)

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> HttpConnector:
        http_config = config.get_interface_by_name("http")
        connector_config = http_config["connector"]

        host = connector_config.get("host")
        port = connector_config.get("port")
        adapter = HttpAdapter.from_config(config)

        return cls(adapter, host, port)

    async def start(self) -> None:
        """Start the connector.

        :return: This method does not return anything.
        """

        logger.info(f"Starting {self!r}...")
        self.mount_routes()
        await self._start()

    async def stop(self) -> None:
        """Stop the connector.

        :return: This method does not return anything.
        """
        logger.info(f"Stopping {self!r}...")
        await self._stop()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError

    def mount_routes(self) -> None:
        """Mount the routes given by the adapter.

        :return: This method does not return anything.
        """
        for decorator, callback in self.routes.items():
            self.mount_route(decorator.path, decorator.method, callback)

    def mount_route(self, path: str, method: str, callback: Callable[[Request], Optional[Response]]):
        """Mount a new route on the application.

        :param path: The request's path.
        :param method: The request's method.
        :param callback: The callback to be executed.
        :return: This method does not return anything.
        """
        adapted_callback = self.adapt_callback(callback)
        self._mount_route(path, method, adapted_callback)

    @abstractmethod
    def _mount_route(self, path: str, method: str, adapted_callback: Callable) -> None:
        raise NotImplementedError

    def adapt_callback(
        self, callback: Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]
    ) -> Callable[[RawRequest], Awaitable[RawResponse]]:
        """Get the adapted callback to be used by the connector.

        :param callback: The function.
        :return: A wrapper function on top of the given one that is compatible with the connector.
        """

        @wraps(callback)
        async def _wrapper(raw: RawRequest) -> RawResponse:
            async with self._semaphore:
                logger.info(f"Dispatching '{raw!s}'...")

                request = await self._build_request(raw)
                token = REQUEST_USER_CONTEXT_VAR.set(request.user)

                # noinspection PyBroadException
                try:
                    response = callback(request)
                    if isawaitable(response):
                        response = await response

                    return await self._build_response(response)

                except ResponseException as exc:
                    tb = traceback.format_exc()
                    logger.error(f"Raised an application exception:\n {tb}")
                    return await self._build_error_response(tb, exc.status)
                except Exception:
                    tb = traceback.format_exc()
                    logger.exception(f"Raised a system exception:\n {tb}")
                    return await self._build_error_response(tb, 500)
                finally:
                    REQUEST_USER_CONTEXT_VAR.reset(token)

        return _wrapper

    @abstractmethod
    async def _build_request(self, request: RawRequest) -> Request:
        raise NotImplementedError

    @abstractmethod
    async def _build_response(self, response: Optional[Response]) -> RawResponse:
        raise NotImplementedError

    @abstractmethod
    async def _build_error_response(self, message: str, status: int) -> RawResponse:
        raise NotImplementedError

    @property
    def host(self) -> str:
        """Get the host.

        :return: A ``str`` value.
        """
        return self._host

    @property
    def port(self) -> int:
        """Get the port.

        :return: A ``int`` value.
        """
        return self._port

    @property
    def routes(self) -> dict[HttpEnrouteDecorator, _Callback]:
        """Get the port.

        :return: A ``int`` value.
        """
        return self.adapter.routes

    @property
    def adapter(self) -> HttpAdapter:
        """Get the port.

        :return: A ``int`` value.
        """
        return self._adapter

    def __repr__(self):
        return f"{type(self).__name__}({self._host!r}, {self._port})"



from .abc import (
    BrokerQueue,
)
from .database import (
    BrokerQueueDatabaseOperationFactory,
    DatabaseBrokerQueue,
    DatabaseBrokerQueueBuilder,
)
from .memory import (
    InMemoryBrokerQueue,
)



from __future__ import (
    annotations,
)

import asyncio
from datetime import (
    datetime,
)
from itertools import (
    count,
)
from math import (
    inf,
)
from typing import (
    Any,
    AsyncIterator,
    Optional,
    Union,
)

from crontab import CronTab as CrontTabImpl

from minos.common import (
    current_datetime,
)


class CronTab:
    """CronTab class."""

    def __init__(self, pattern: Union[str, CrontTabImpl]):
        if isinstance(pattern, str) and pattern == "@reboot":
            pattern = None
        elif not isinstance(pattern, CrontTabImpl):
            pattern = CrontTabImpl(pattern)
        self._impl = pattern

    @property
    def impl(self) -> Optional[CrontTabImpl]:
        """Get the crontab implementation.

        :return: A ``crontab.CronTab`` or ``None``.
        """
        return self._impl

    def __hash__(self):
        return hash((type(self), self._impl_matchers))

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self._impl_matchers == other._impl_matchers

    @property
    def _impl_matchers(self):
        if self._impl is None:
            return None
        return self._impl.matchers

    async def __aiter__(self) -> AsyncIterator[datetime]:
        counter = count()
        now = current_datetime()
        while next(counter) < self.repetitions:
            await self.sleep_until_next(now)
            now = current_datetime()
            yield now

        await asyncio.sleep(inf)

    async def sleep_until_next(self, *args, **kwargs) -> None:
        """Sleep until next matching.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        await asyncio.sleep(self.get_delay_until_next(*args, **kwargs))

    def get_delay_until_next(self, now: Optional[datetime] = None) -> float:
        """Get the time to wait for next matching.

        :param now: Current time.
        :return:
        """
        if self._impl is None:
            return 0

        if now is None:
            now = current_datetime()
        return self._impl.next(now)

    @property
    def repetitions(self) -> Union[int, float]:
        """Get the number of repetitions.

        :return: A ``float`` value.
        """
        if self._impl is None:
            return 1
        return inf



from .abc import (
    BrokerSubscriberValidator,
)
from .duplicates import (
    BrokerSubscriberDuplicateValidator,
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
    InMemoryBrokerSubscriberDuplicateValidator,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
)
from collections.abc import (
    Iterable,
)
from typing import (
    TYPE_CHECKING,
    Final,
    Union,
)

from .abc import (
    EnrouteDecorator,
)
from .kinds import (
    EnrouteDecoratorKind,
)

if TYPE_CHECKING:
    from crontab import CronTab as CronTabImpl

    from ...scheduling import (
        CronTab,
    )


class PeriodicEnrouteDecorator(EnrouteDecorator, ABC):
    """Periodic Enroute class"""

    def __init__(self, crontab: Union[str, CronTab, CronTabImpl], **kwargs):
        super().__init__(**kwargs)
        from ...scheduling import (
            CronTab,
        )

        if not isinstance(crontab, CronTab):
            crontab = CronTab(crontab)
        self.crontab = crontab

    def __iter__(self) -> Iterable:
        yield from (self.crontab,)


class PeriodicEventEnrouteDecorator(PeriodicEnrouteDecorator):
    """Periodic Event Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Event



from .api import (
    enroute,
)
from .callables import (
    Checker,
    CheckerMeta,
    CheckerWrapper,
    Handler,
    HandlerMeta,
    HandlerWrapper,
)
from .collectors import (
    EnrouteCollector,
)
from .definitions import (
    BrokerCommandEnrouteDecorator,
    BrokerEnrouteDecorator,
    BrokerEventEnrouteDecorator,
    BrokerQueryEnrouteDecorator,
    CheckDecorator,
    EnrouteDecorator,
    EnrouteDecoratorKind,
    HttpEnrouteDecorator,
    PeriodicEnrouteDecorator,
    PeriodicEventEnrouteDecorator,
    RestCommandEnrouteDecorator,
    RestEnrouteDecorator,
    RestQueryEnrouteDecorator,
)
from .factories import (
    EnrouteFactory,
)



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
)

from ....collections import (
    BrokerQueueDatabaseOperationFactory,
    DatabaseBrokerQueue,
)
from .abc import (
    BrokerPublisherQueue,
)

logger = logging.getLogger(__name__)


class BrokerPublisherQueueDatabaseOperationFactory(BrokerQueueDatabaseOperationFactory, ABC):
    """Broker Publisher Queue Database Operation Factory class."""


class DatabaseBrokerPublisherQueue(
    DatabaseBrokerQueue[BrokerPublisherQueueDatabaseOperationFactory], BrokerPublisherQueue
):
    """Database Broker Publisher Queue class."""



from .brokers import (
    MockedBrokerPublisherQueueDatabaseOperationFactory,
    MockedBrokerQueueDatabaseOperationFactory,
    MockedBrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    MockedBrokerSubscriberQueueDatabaseOperationFactory,
)



from typing import (
    Final,
)

from ..kinds import (
    EnrouteDecoratorKind,
)
from .abc import (
    HttpEnrouteDecorator,
)


class RestEnrouteDecorator(HttpEnrouteDecorator):
    """Rest Enroute class"""


class RestCommandEnrouteDecorator(RestEnrouteDecorator):
    """Rest Command Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Command


class RestQueryEnrouteDecorator(RestEnrouteDecorator):
    """Rest Query Enroute class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Query



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from minos.common import (
    BuildableMixin,
    Builder,
    Config,
    Injectable,
    MinosConfigException,
)

from ..messages import (
    BrokerMessage,
)

if TYPE_CHECKING:
    from .queued import (
        BrokerPublisherQueue,
        QueuedBrokerPublisher,
    )

logger = logging.getLogger(__name__)


@Injectable("broker_publisher")
class BrokerPublisher(ABC, BuildableMixin):
    """Broker Publisher class."""

    async def send(self, message: BrokerMessage) -> None:
        """Send a message.

        :param message: The message to be sent.
        :return: This method does not return anything.
        """
        logger.debug(f"Sending {message!r} message...")
        await self._send(message)

    @abstractmethod
    async def _send(self, message: BrokerMessage) -> None:
        raise NotImplementedError


BrokerPublisherCls = TypeVar("BrokerPublisherCls", bound=BrokerPublisher)


class BrokerPublisherBuilder(Builder[BrokerPublisher], Generic[BrokerPublisherCls]):
    """Broker Publisher Builder class."""

    def __init__(
        self,
        *args,
        queue_builder: Optional[Builder] = None,
        queued_cls: Optional[type[QueuedBrokerPublisher]] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if queued_cls is None:
            from .queued import (
                QueuedBrokerPublisher,
            )

            queued_cls = QueuedBrokerPublisher

        self.queue_builder = queue_builder

        self.queued_cls = queued_cls

    def with_queued_cls(self, queued_cls: type[QueuedBrokerPublisher]):
        """Set the queued class.

        :param queued_cls: A subclass of ``QueuedBrokerPublisher``.
        :return: This method return the builder instance.
        """
        self.queued_cls = queued_cls

        return self

    def with_config(self, config: Config):
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self._with_builders_from_config(config)

        if self.queue_builder is not None:
            self.queue_builder.with_config(config)
        return super().with_config(config)

    def _with_builders_from_config(self, config):
        try:
            broker_config = config.get_interface_by_name("broker")
        except MinosConfigException:
            return

        broker_publisher_config = broker_config["publisher"]

        if "queue" in broker_publisher_config:
            self.with_queue(broker_publisher_config["queue"])

    def with_queue(self, queue: Union[type[BrokerPublisherQueue], Builder[BrokerPublisherQueue]]):
        """Set the queue builder.

        :param queue: The queue builder to be set.
        :return: This method return the builder instance.
        """
        if not isinstance(queue, Builder):
            queue = queue.get_builder()
        self.queue_builder = queue.copy()
        return self

    def with_kwargs(self, kwargs: dict[str, Any]):
        """Set kwargs.

        :param kwargs: The kwargs to be set.
        :return: This method return the builder instance.
        """
        if self.queue_builder is not None:
            self.queue_builder.with_kwargs(kwargs)

        return super().with_kwargs(kwargs)

    def build(self) -> BrokerPublisher:
        """Build the instance.

        :return: A ``QueuedBrokerSubscriber`` instance.
        """
        impl = super().build()

        if self.queue_builder is not None:
            queue = self.queue_builder.build()
            impl = self.queued_cls(impl=impl, queue=queue, **self.kwargs)

        return impl


BrokerPublisher.set_builder(BrokerPublisherBuilder)



from .impl import (
    QueuedBrokerPublisher,
)
from .queues import (
    BrokerPublisherQueue,
    BrokerPublisherQueueDatabaseOperationFactory,
    DatabaseBrokerPublisherQueue,
    InMemoryBrokerPublisherQueue,
)



from ..decorators import (
    enroute,
)
from ..requests import (
    Request,
    Response,
)
from ..utils import (
    get_host_ip,
)


class SystemService:
    """System Service class."""

    # noinspection PyUnusedLocal
    @enroute.rest.command("/system/health", "GET")
    def check_health(self, request: Request) -> Response:
        """Get the system health.

        :param request: The given request.
        :return: A Response containing the system status.
        """
        return Response({"host": get_host_ip()})



from abc import (
    ABC,
    abstractmethod,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


# noinspection SqlNoDataSourceInspection,SqlResolve
class BrokerSubscriberDuplicateValidatorDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Broker Subscriber Duplicate Validator Database Operation Factory class."""

    @abstractmethod
    def build_create(self) -> DatabaseOperation:
        """Build the "create table" query.

        :return: A ``SQL`` instance.
        """

    @abstractmethod
    def build_submit(self, topic: str, uuid: UUID) -> DatabaseOperation:
        """Build the "insert row" query.

        :return: A ``SQL`` instance.
        """



from .adapters import (
    HttpAdapter,
)
from .connectors import (
    HttpConnector,
)
from .ports import (
    HttpPort,
    RestService,
)
from .requests import (
    HttpRequest,
    HttpResponse,
    HttpResponseException,
)



from .crontab import (
    CronTab,
)
from .ports import (
    PeriodicPort,
    PeriodicTaskSchedulerService,
)
from .requests import (
    ScheduledRequest,
    ScheduledRequestContent,
    ScheduledResponseException,
)
from .schedulers import (
    PeriodicTask,
    PeriodicTaskScheduler,
)



from .factories import (
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
)
from .impl import (
    DatabaseBrokerSubscriberDuplicateValidator,
    DatabaseBrokerSubscriberDuplicateValidatorBuilder,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Iterable,
)
from typing import (
    Final,
    Union,
)

from ..callables import (
    Handler,
    HandlerMeta,
    HandlerWrapper,
)
from .kinds import (
    EnrouteDecoratorKind,
)


class EnrouteDecorator(ABC):
    """Base Decorator class."""

    # noinspection PyFinal
    KIND: Final[EnrouteDecoratorKind]

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, func: Union[Handler, HandlerWrapper]) -> HandlerWrapper:
        if isinstance(func, HandlerWrapper):
            meta = func.meta
        else:
            meta = HandlerMeta(func)

        meta.add_decorator(self)

        return meta.wrapper

    def __repr__(self):
        args = ", ".join(map(repr, self))
        return f"{type(self).__name__}({args})"

    def __eq__(self, other: EnrouteDecorator) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    @abstractmethod
    def __iter__(self) -> Iterable:
        raise NotImplementedError

    @property
    def pre_fn_name(self) -> str:
        """Get the pre execution function name.

        :return: A string value containing the function name.
        """
        return self.KIND.pre_fn_name

    @property
    def post_fn_name(self) -> str:
        """Get the post execution function name.

        :return: A string value containing the function name.
        """
        return self.KIND.post_fn_name



from __future__ import (
    annotations,
)

from collections.abc import (
    Awaitable,
    Callable,
)
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from .abc import (
    Request,
)

sentinel = object()
ContentAction = Callable[[Any, ...], Union[Any, Awaitable[Any]]]
ParamsAction = Callable[[dict[str, Any], ...], Union[dict[str, Any], Awaitable[dict[str, Any]]]]


class WrappedRequest(Request):
    """Wrapped Request class."""

    def __init__(
        self,
        base: Request,
        content_action: Optional[ContentAction] = None,
        params_action: Optional[ParamsAction] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.base = base
        self.content_action = content_action
        self.params_action = params_action
        self._computed_content = sentinel
        self._computed_params = sentinel

    @property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        return self.base.user

    @property
    def has_content(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        return self.base.has_content

    async def _content(self, **kwargs) -> Any:
        if self.content_action is None:
            return await self.base.content()

        if self._computed_content is sentinel:
            content = self.content_action(await self.base.content(), **kwargs)
            if isawaitable(content):
                content = await content
            self._computed_content = content
        return self._computed_content

    @property
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        return self.base.has_params

    async def _params(self, **kwargs) -> Any:
        if self.params_action is None:
            return await self.base.params()

        if self._computed_params is sentinel:
            params = self.params_action(await self.base.params(), **kwargs)
            if isawaitable(params):
                params = await params
            self._computed_params = params
        return self._computed_params

    def __eq__(self, other: WrappedRequest) -> bool:
        return (
            isinstance(other, type(self))
            and self.base == other.base
            and self.content_action == other.content_action
            and self.params_action == other.params_action
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.base!r}, {self.content_action!r}, {self.params_action!r})"



import logging
import warnings

from cached_property import (
    cached_property,
)

from minos.common import (
    Port,
)

from .schedulers import (
    PeriodicTaskScheduler,
)

logger = logging.getLogger(__name__)


class PeriodicPort(Port):
    """Periodic Port class."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_kwargs = kwargs

    async def _start(self) -> None:
        await self.scheduler.setup()

        try:
            self.start_event.set()
        except RuntimeError:
            logger.warning("Runtime is not properly setup.")

        await self.scheduler.start()

    async def _stop(self, exception: Exception = None) -> None:
        await self.scheduler.stop()
        await self.scheduler.destroy()

    @cached_property
    def scheduler(self) -> PeriodicTaskScheduler:
        """Get the service scheduler.

        :return: A ``PeriodicTaskScheduler`` instance.
        """
        return PeriodicTaskScheduler.from_config(**self._init_kwargs)


class PeriodicTaskSchedulerService(PeriodicPort):
    """Periodic Task Scheduler Service class."""

    def __init__(self, **kwargs):
        warnings.warn(
            f"{PeriodicTaskSchedulerService!r} has been deprecated. Use {PeriodicPort} instead.",
            DeprecationWarning,
        )
        super().__init__(**kwargs)



from itertools import (
    chain,
)

from minos.common import (
    Config,
)

from ..decorators import (
    EnrouteCollector,
    enroute,
)
from ..requests import (
    Request,
    Response,
)


class AsyncAPIService:
    def __init__(self, config: Config):
        self.config = config
        self.spec = SPECIFICATION_SCHEMA.copy()

    @enroute.rest.command("/spec/asyncapi", "GET")
    def generate_specification(self, request: Request) -> Response:
        events = self.get_events()

        for event in events:
            topic: str = event["topic"]
            event_spec = {}

            self.spec["channels"][topic] = event_spec

        return Response(self.spec)

    def get_events(self) -> list[dict]:
        events = list()
        for name in self.config.get_services():
            decorators = EnrouteCollector(name, self.config).get_broker_event()
            events += [{"topic": decorator.topic} for decorator in set(chain(*decorators.values()))]

        return events


SPECIFICATION_SCHEMA = {
    "asyncapi": "2.3.0",
    "info": {"title": "", "version": ""},
    "channels": {},
}



from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
)

from minos.common import (
    DatabaseClient,
)

from .....collections import (
    DatabaseBrokerQueue,
    DatabaseBrokerQueueBuilder,
)
from ..abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)
from .factories import (
    BrokerSubscriberQueueDatabaseOperationFactory,
)

logger = logging.getLogger(__name__)


class DatabaseBrokerSubscriberQueue(
    DatabaseBrokerQueue[BrokerSubscriberQueueDatabaseOperationFactory], BrokerSubscriberQueue
):
    """Database Broker Subscriber Queue class."""

    async def _get_count(self) -> int:
        # noinspection PyTypeChecker
        operation = self.database_operation_factory.build_count(self._retry, self.topics)
        row = await self.execute_on_database_and_fetch_one(operation)
        count = row[0]
        return count

    async def _dequeue_rows(self, client: DatabaseClient) -> list[Any]:
        operation = self.database_operation_factory.build_query(self._retry, self._records, self.topics)
        await client.execute(operation)
        return [row async for row in client.fetch_all()]


class DatabaseBrokerSubscriberQueueBuilder(
    BrokerSubscriberQueueBuilder[DatabaseBrokerSubscriberQueue], DatabaseBrokerQueueBuilder
):
    """Database Broker Subscriber Queue Builder class."""


DatabaseBrokerSubscriberQueue.set_builder(DatabaseBrokerSubscriberQueueBuilder)



import logging

from ....collections import (
    InMemoryBrokerQueue,
)
from .abc import (
    BrokerPublisherQueue,
)

logger = logging.getLogger(__name__)


class InMemoryBrokerPublisherQueue(InMemoryBrokerQueue, BrokerPublisherQueue):
    """In Memory Broker Publisher Queue class."""



import logging
from asyncio import (
    Queue,
    QueueEmpty,
    TimeoutError,
    wait_for,
)

from ...messages import (
    BrokerMessage,
)
from .abc import (
    BrokerQueue,
)

logger = logging.getLogger(__name__)


class InMemoryBrokerQueue(BrokerQueue):
    """In Memory Broker Queue class."""

    _queue: Queue[BrokerMessage]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = Queue()

    async def _destroy(self) -> None:
        await self._flush_queue()
        await super()._destroy()

    async def _flush_queue(self):
        try:
            await wait_for(self._queue.join(), 0.5)
        except TimeoutError:
            messages = list()
            while True:
                try:
                    messages.append(self._queue.get_nowait())
                except QueueEmpty:
                    break
            logger.warning(f"Some messages were loosed: {messages}")

    async def _enqueue(self, message: BrokerMessage) -> None:
        await self._queue.put(message)

    async def _dequeue(self) -> BrokerMessage:
        message = await self._queue.get()
        self._queue.task_done()
        return message



from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
)
from minos.common.testing import (
    MockedDatabaseClient,
    MockedDatabaseOperation,
)

from ....brokers import (
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
)


class MockedBrokerSubscriberDuplicateValidatorDatabaseOperationFactory(
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory
):
    """For testing purposes"""

    def build_create(self) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create_table")

    def build_submit(self, topic: str, uuid: UUID) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("insert_row")


MockedDatabaseClient.set_factory(
    BrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
    MockedBrokerSubscriberDuplicateValidatorDatabaseOperationFactory,
)



from itertools import (
    chain,
)
from operator import (
    itemgetter,
)

from minos.common import (
    Config,
)

from ..decorators import (
    EnrouteCollector,
    enroute,
)
from ..requests import (
    Request,
    Response,
)


class OpenAPIService:
    def __init__(self, config: Config):
        self.config = config
        self.spec = SPECIFICATION_SCHEMA.copy()

    # noinspection PyUnusedLocal
    @enroute.rest.command("/spec/openapi", "GET")
    def generate_specification(self, request: Request) -> Response:
        for endpoint in self.endpoints:
            url = endpoint["url"]
            method = endpoint["method"].lower()

            if url in self.spec["paths"]:
                self.spec["paths"][url][method] = PATH_SCHEMA
            else:
                self.spec["paths"][url] = {method: PATH_SCHEMA}

        return Response(self.spec)

    @property
    def endpoints(self) -> list[dict]:
        endpoints = list()
        for name in self.config.get_services():
            decorators = EnrouteCollector(name, self.config).get_rest_command_query()
            endpoints += [
                {"url": decorator.url, "method": decorator.method} for decorator in set(chain(*decorators.values()))
            ]

        endpoints.sort(key=itemgetter("url", "method"))

        return endpoints


SPECIFICATION_SCHEMA = {
    "openapi": "3.0.0",
    "info": {
        "version": "1.0.0",
        "title": "Minos OpenAPI Spec",
        "description": "Minos OpenAPI Spec",
    },
    "paths": {},
}

PATH_SCHEMA = {
    "responses": {"200": {"description": ""}},
}



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from uuid import (
    UUID,
)

from .....messages import (
    BrokerMessage,
)
from ..abc import (
    BrokerSubscriberValidator,
)


class BrokerSubscriberDuplicateValidator(BrokerSubscriberValidator, ABC):
    """Broker Subscriber Duplicate Detector class."""

    async def _is_valid(self, message: BrokerMessage) -> bool:
        """Check if the given message is valid.

        :param message: The message to be checked.
        :return: ``True`` if it is valid or ``False`` otherwise.
        """
        return await self._is_unique(message.topic, message.identifier)

    @abstractmethod
    async def _is_unique(self, topic: str, uuid: UUID) -> bool:
        raise NotImplementedError



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Callable,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Config,
    SetupMixin,
)

from .decorators import (
    BrokerEnrouteDecorator,
    EnrouteDecorator,
    EnrouteFactory,
    HttpEnrouteDecorator,
    PeriodicEnrouteDecorator,
)


class Router(ABC, SetupMixin):
    """Router base class."""

    def __init__(self, config: Config, **kwargs):
        super().__init__(**kwargs)

        self._config = config

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> Router:
        return cls(config)

    @cached_property
    def routes(self) -> dict[EnrouteDecorator, Callable]:
        """Get the routes stored on the router.

        :return: A dict with decorators as keys and callbacks as values.
        """
        return self._build_routes()

    def _build_routes(self) -> dict[EnrouteDecorator, Callable]:
        routes = self._get_all_routes()
        routes = self._filter_routes(routes)
        return routes

    def _get_all_routes(self) -> dict[EnrouteDecorator, Callable]:
        builder = EnrouteFactory(*self._config.get_services(), middleware=self._config.get_middleware())
        routes = builder.get_all(config=self._config)
        return routes

    @abstractmethod
    def _filter_routes(self, routes: dict[EnrouteDecorator, Callable]) -> dict[EnrouteDecorator, Callable]:
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if self.routes.keys() != other.routes.keys():
            return False

        for key in self.routes:
            if self.routes[key].__qualname__ != other.routes[key].__qualname__:
                return False

        return True


class HttpRouter(Router, ABC):
    """Http Router base class."""

    routes: dict[HttpEnrouteDecorator, Callable]


class RestHttpRouter(HttpRouter):
    """Rest Http Router class."""

    def _filter_routes(self, routes: dict[EnrouteDecorator, Callable]) -> dict[EnrouteDecorator, Callable]:
        routes = {
            decorator: callback for decorator, callback in routes.items() if isinstance(decorator, HttpEnrouteDecorator)
        }
        return routes


class BrokerRouter(Router):
    """Broker Router class."""

    def _filter_routes(self, routes: dict[EnrouteDecorator, Callable]) -> dict[EnrouteDecorator, Callable]:
        routes = {
            decorator: callback
            for decorator, callback in routes.items()
            if isinstance(decorator, BrokerEnrouteDecorator)
        }
        return routes


class PeriodicRouter(Router):
    """Periodic Router class."""

    def _filter_routes(self, routes: dict[EnrouteDecorator, Callable]) -> dict[EnrouteDecorator, Callable]:
        routes = {
            decorator: callback
            for decorator, callback in routes.items()
            if isinstance(decorator, PeriodicEnrouteDecorator)
        }
        return routes
